@@ -3,6 +3,7 @@
 
 import itertools
 import logging
+import typing as T
 
 from . import enums, errors, utils
 
@@ -23,16 +24,28 @@ class BaseDeclaration(utils.OrderedBase):
     #: Set to False on declarations that perform their own unrolling.
     UNROLL_CONTEXT_BEFORE_EVALUATION = True
 
+    def __init__(self, **defaults):
+        super().__init__()
+        self._defaults = defaults or {}
+
     def unroll_context(self, instance, step, context):
+        full_context = dict()
+        full_context.update(self._defaults)
+        full_context.update(context)
+
         if not self.UNROLL_CONTEXT_BEFORE_EVALUATION:
-            return context
-        if not any(enums.get_builder_phase(v) for v in context.values()):
+            return full_context
+        if not any(enums.get_builder_phase(v) for v in full_context.values()):
             # Optimization for simple contexts - don't do anything.
-            return context
+            return full_context
 
         import factory.base
         subfactory = factory.base.DictFactory
-        return step.recurse(subfactory, context, force_sequence=step.sequence)
+        return step.recurse(subfactory, full_context, force_sequence=step.sequence)
+
+    def evaluate_pre(self, instance, step, overrides):
+        context = self.unroll_context(instance, step, overrides)
+        return self.evaluate(instance, step, context)
 
     def evaluate(self, instance, step, extra):
         """Evaluate this declaration.
@@ -61,8 +74,8 @@ class LazyFunction(BaseDeclaration):
             returning the computed value.
     """
 
-    def __init__(self, function, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, function):
+        super().__init__()
         self.function = function
 
     def evaluate(self, instance, step, extra):
@@ -78,8 +91,8 @@ class LazyAttribute(BaseDeclaration):
             returning the computed value.
     """
 
-    def __init__(self, function, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, function):
+        super().__init__()
         self.function = function
 
     def evaluate(self, instance, step, extra):
@@ -133,8 +146,8 @@ class SelfAttribute(BaseDeclaration):
             exist.
     """
 
-    def __init__(self, attribute_name, default=_UNSPECIFIED, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, attribute_name, default=_UNSPECIFIED):
+        super().__init__()
         depth = len(attribute_name) - len(attribute_name.lstrip('.'))
         attribute_name = attribute_name[depth:]
 
@@ -241,8 +254,8 @@ class ContainerAttribute(BaseDeclaration):
         strict (bool): Whether evaluating should fail when the containers are
             not passed in (i.e used outside a SubFactory).
     """
-    def __init__(self, function, strict=True, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, function, strict=True):
+        super().__init__()
         self.function = function
         self.strict = strict
 
@@ -272,28 +285,8 @@ class ParameteredAttribute(BaseDeclaration):
     Attributes:
         defaults (dict): Default values for the parameters.
             May be overridden by call-time parameters.
-
-    Class attributes:
-        CONTAINERS_FIELD (str): name of the field, if any, where container
-            information (e.g for SubFactory) should be stored. If empty,
-            containers data isn't merged into generate() parameters.
     """
 
-    CONTAINERS_FIELD = '__containers'
-
-    # Whether to add the current object to the stack of containers
-    EXTEND_CONTAINERS = False
-
-    def __init__(self, **kwargs):
-        super().__init__()
-        self.defaults = kwargs
-
-    def _prepare_containers(self, obj, containers=()):
-        if self.EXTEND_CONTAINERS:
-            return (obj,) + tuple(containers)
-
-        return containers
-
     def evaluate(self, instance, step, extra):
         """Evaluate the current definition and fill its attributes.
 
@@ -308,11 +301,7 @@ class ParameteredAttribute(BaseDeclaration):
             extra (dict): additional, call-time added kwargs
                 for the step.
         """
-        defaults = dict(self.defaults)
-        if extra:
-            defaults.update(extra)
-
-        return self.generate(step, defaults)
+        return self.generate(step, extra)
 
     def generate(self, step, params):
         """Actually generate the related attribute.
@@ -331,39 +320,6 @@ class ParameteredAttribute(BaseDeclaration):
         raise NotImplementedError()
 
 
-class ParameteredDeclaration(BaseDeclaration):
-    """A declaration with parameters.
-
-    The parameters can be any factory-enabled declaration, and will be resolved
-    before the call to the user-defined code in `self.generate()`.
-
-    Attributes:
-        defaults (dict): Default values for the parameters; can be overridden
-            by call-time parameters. Accepts BaseDeclaration subclasses.
-    """
-
-    def __init__(self, **defaults):
-        self.defaults = defaults
-        super().__init__()
-
-    def unroll_context(self, instance, step, context):
-        merged_context = {}
-        merged_context.update(self.defaults)
-        merged_context.update(context)
-        return super().unroll_context(instance, step, merged_context)
-
-    def evaluate(self, instance, step, extra):
-        return self.generate(extra)
-
-    def generate(self, params):
-        """Generate a value for this declaration.
-
-        Args:
-            params (dict): the parameters, after a factory evaluation.
-        """
-        raise NotImplementedError()
-
-
 class _FactoryWrapper:
     """Handle a 'factory' arg.
 
@@ -393,12 +349,12 @@ class _FactoryWrapper:
 
     def __repr__(self):
         if self.factory is None:
-            return '<_FactoryImport: %s.%s>' % (self.module, self.name)
+            return f'<_FactoryImport: {self.module}.{self.name}>'
         else:
-            return '<_FactoryImport: %s>' % self.factory.__class__
+            return f'<_FactoryImport: {self.factory.__class__}>'
 
 
-class SubFactory(ParameteredAttribute):
+class SubFactory(BaseDeclaration):
     """Base class for attributes based upon a sub-factory.
 
     Attributes:
@@ -407,7 +363,6 @@ class SubFactory(ParameteredAttribute):
         factory (base.Factory): the wrapped factory
     """
 
-    EXTEND_CONTAINERS = True
     # Whether to align the attribute's sequence counter to the holding
     # factory's sequence counter
     FORCE_SEQUENCE = False
@@ -421,7 +376,7 @@ class SubFactory(ParameteredAttribute):
         """Retrieve the wrapped factory.Factory subclass."""
         return self.factory_wrapper.get()
 
-    def generate(self, step, params):
+    def evaluate(self, instance, step, extra):
         """Evaluate the current definition and fill its attributes.
 
         Args:
@@ -433,11 +388,11 @@ class SubFactory(ParameteredAttribute):
         logger.debug(
             "SubFactory: Instantiating %s.%s(%s), create=%r",
             subfactory.__module__, subfactory.__name__,
-            utils.log_pprint(kwargs=params),
+            utils.log_pprint(kwargs=extra),
             step,
         )
         force_sequence = step.sequence if self.FORCE_SEQUENCE else None
-        return step.recurse(subfactory, params, force_sequence=force_sequence)
+        return step.recurse(subfactory, extra, force_sequence=force_sequence)
 
 
 class Dict(SubFactory):
@@ -490,47 +445,50 @@ class Maybe(BaseDeclaration):
         used_phases = {phase for phase in phases.values() if phase is not None}
 
         if len(used_phases) > 1:
-            raise TypeError("Inconsistent phases for %r: %r" % (self, phases))
+            raise TypeError(f"Inconsistent phases for {self!r}: {phases!r}")
 
         self.FACTORY_BUILDER_PHASE = used_phases.pop() if used_phases else enums.BuilderPhase.ATTRIBUTE_RESOLUTION
 
-    def call(self, instance, step, context):
+    def evaluate_post(self, instance, step, overrides):
+        """Handle post-generation declarations"""
         decider_phase = enums.get_builder_phase(self.decider)
         if decider_phase == enums.BuilderPhase.ATTRIBUTE_RESOLUTION:
             # Note: we work on the *builder stub*, not on the actual instance.
             # This gives us access to all Params-level definitions.
-            choice = self.decider.evaluate(instance=step.stub, step=step, extra=context.extra)
+            choice = self.decider.evaluate_pre(
+                instance=step.stub, step=step, overrides=overrides)
         else:
             assert decider_phase == enums.BuilderPhase.POST_INSTANTIATION
-            choice = self.decider.call(instance, step, context)
+            choice = self.decider.evaluate_post(
+                instance=instance, step=step, overrides={})
 
         target = self.yes if choice else self.no
         if enums.get_builder_phase(target) == enums.BuilderPhase.POST_INSTANTIATION:
-            return target.call(
+            return target.evaluate_post(
                 instance=instance,
                 step=step,
-                context=context,
+                overrides=overrides,
             )
         else:
             # Flat value (can't be ATTRIBUTE_RESOLUTION, checked in __init__)
             return target
 
-    def evaluate(self, instance, step, extra):
+    def evaluate_pre(self, instance, step, overrides):
         choice = self.decider.evaluate(instance=instance, step=step, extra={})
         target = self.yes if choice else self.no
 
         if isinstance(target, BaseDeclaration):
-            return target.evaluate(
+            return target.evaluate_pre(
                 instance=instance,
                 step=step,
-                extra=extra,
+                overrides=overrides,
             )
         else:
             # Flat value (can't be POST_INSTANTIATION, checked in __init__)
             return target
 
     def __repr__(self):
-        return 'Maybe(%r, yes=%r, no=%r)' % (self.decider, self.yes, self.no)
+        return f'Maybe({self.decider!r}, yes={self.yes!r}, no={self.no!r})'
 
 
 class Parameter(utils.OrderedBase):
@@ -613,11 +571,26 @@ class Trait(Parameter):
 # ===============
 
 
+class PostGenerationContext(T.NamedTuple):
+    value_provided: bool
+    value: T.Any
+    extra: T.Dict[str, T.Any]
+
+
 class PostGenerationDeclaration(BaseDeclaration):
     """Declarations to be called once the model object has been generated."""
 
     FACTORY_BUILDER_PHASE = enums.BuilderPhase.POST_INSTANTIATION
 
+    def evaluate_post(self, instance, step, overrides):
+        context = self.unroll_context(instance, step, overrides)
+        postgen_context = PostGenerationContext(
+            value_provided=bool('' in context),
+            value=context.get(''),
+            extra={k: v for k, v in context.items() if k != ''},
+        )
+        return self.call(instance, step, postgen_context)
+
     def call(self, instance, step, context):  # pragma: no cover
         """Call this hook; no return value is expected.
 
@@ -717,9 +690,11 @@ class RelatedFactoryList(RelatedFactory):
         super().__init__(factory, factory_related_name, **defaults)
 
     def call(self, instance, step, context):
-        return [super(RelatedFactoryList, self).call(instance, step, context)
-                for i in range(self.size if isinstance(self.size, int)
-                               else self.size())]
+        parent = super()
+        return [
+            parent.call(instance, step, context)
+            for i in range(self.size if isinstance(self.size, int) else self.size())
+        ]
 
 
 class NotProvided:
